import argparse
from typing import cast

from pyspark.sql import SparkSession
from pyspark.sql.functions import lit, when, length, col, format_string
from datasets import Dataset, Features, Value  # type: ignore
from transformers import AutoTokenizer, DataCollatorWithPadding, TFAutoModelForSequenceClassification  # type: ignore
import tensorflow as tf  # type: ignore
import pandas as pd
import trend_etl.schema.args as arguments

program = "classify_new_subreddits"
database = "redditpl"
collection = "new_subreddits"
model_path = "/dbfs/FileStore/subreddit_expansion/models/distilbert-uncased-8/"
id_key = "upsert_field"
threshold_prob = (
    0.876  # subreddits above this probability was giving good results with this model
)


# Below code can be used to save and load tokenizer:
#
# from transformers import AutoTokenizer, AutoConfig
# tokenizer = AutoTokenizer.from_pretrained('roberta-base')
# tokenizer.save_pretrained('TEST/tokenizer')
# tokenizer = AutoTokenizer.from_pretrained('TEST/tokenizer', config=AutoConfig.from_pretrained("roberta-base"))
def prep_df_for_prediction(df, model):
    """Prepare a dataframe of subreddits for prediction.
    `df` is assumed to be a pandas dataframe with 'subreddit'
    as the index, and with (string) columns for:
    - title
    - advertiser_category
    - public_description
    - description
    - submit_text
    `model` is the model the dataframe is being prepared for."""
    # Create a Dataset object from the dataframe
    features = Features({"text": Value("string")})
    subs = Dataset.from_spark(
        df, features=features, cache_dir="/dbfs/FileStore/classify_subreddits_model"
    )

    # Tokenize the input text
    tokenizer = AutoTokenizer.from_pretrained("distilbert-base-uncased")

    def preprocess(examples):
        return tokenizer(examples["text"], truncation=True)

    tokenized_subs = subs.map(preprocess, batched=True)

    data_collator = DataCollatorWithPadding(tokenizer=tokenizer, return_tensors="tf")
    return model.prepare_tf_dataset(
        tokenized_subs, shuffle=False, batch_size=16, collate_fn=data_collator
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=[
            arguments.catalogs[arguments.catalogs_dev],
            arguments.catalogs[arguments.catalogs_prod],
        ],
        help="The unit catalog to use",
    )
    args = parser.parse_args()

    if args.target_catalog == arguments.catalogs[arguments.catalogs_dev]:
        return  # do not run dev version of this job

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    new_subreddits = (
        spark.read.format("mongodb")
        .option("database", database)
        .option("collection", collection)
        .load()
        .drop("_id", "prob")
        .where(col("classification_status") == "new")
        .cache()
    )

    num_new = new_subreddits.count()
    print("New subreddits", num_new)

    if num_new > 0:
        new_subreddits_text = new_subreddits.withColumn(
            "text",
            format_string(
                "<name> %s <title> %s <ad_cat> %s <pub_desc> %s <description> %s <sub> %s",
                col("name"),
                col("data.title"),
                col("data.advertiser_category"),
                col("data.public_description"),
                col("data.description"),
                col("data.submit_text"),
            ),
        )

        mymodel = TFAutoModelForSequenceClassification.from_pretrained(model_path)

        data = prep_df_for_prediction(new_subreddits_text.selectExpr("text"), mymodel)

        logits = mymodel.predict(data).logits
        probs = tf.nn.softmax(logits)[:, 1]

        pandas = new_subreddits_text.select(id_key).toPandas()
        pandas["prob"] = probs

        result_df = spark.createDataFrame(pandas)

        output_df = new_subreddits.join(result_df, id_key, "inner")

        (
            output_df.withColumn(
                "classification_status",
                when((col("prob") > threshold_prob), lit("add")).otherwise(
                    lit("ignore")
                ),
            )
            .write.format("mongodb")
            .option("database", database)
            .option("collection", collection)
            .option("idFieldList", id_key)
            .option("operationType", "replace")
            .option("upsertDocument", "false")
            .mode("append")
            .save()
        )


if __name__ == "__main__":
    main()
