import argparse
from typing import cast

from pyspark.sql import SparkSession
from pyspark.dbutils import DBUtils  # type: ignore
from pyspark.sql.functions import to_json, col

from trend_etl.functions.index import full_table_name
from trend_etl.transform.guilds import from_kafka_get_bronze_guilds_df
import trend_etl.schema.args as arguments

program = "dump_discord_metadata"

postgres_host = "trend-prodogy-instance.cs7lhsrxph4x.us-east-1.rds.amazonaws.com"
postgres_db = "connect"
postgres_port = 5432
postgres_user = "databricks"


def get_postgres_password(spark: SparkSession) -> str:
    dbutils = DBUtils(spark)

    token = dbutils.secrets.get(scope="postgres", key="password")

    return token


def overwrite_guilds_table(spark: SparkSession, catalog: str, postgres_password: str):
    guilds = full_table_name(
        catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_guilds],
    )
    (
        spark.table(guilds)
        .withColumn("source_keys", to_json("source_keys"))
        .write.format("postgresql")
        .mode("overwrite")
        .option("truncate", "true")
        .option("dbtable", "databricks_discord.guilds")
        .option("host", postgres_host)
        .option("port", postgres_port)
        .option("database", postgres_db)
        .option("user", postgres_user)
        .option("password", postgres_password)
        .save()
    )


def overwrite_channels_table(spark: SparkSession, catalog: str, postgres_password: str):
    channels = full_table_name(
        catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_channels],
    )
    (
        spark.table(channels)
        .withColumn("permission_overwrites", to_json("permission_overwrites"))
        .withColumn("recipients", to_json("recipients"))
        .withColumn("thread_metadata", to_json("thread_metadata"))
        .withColumn("member", to_json("member"))
        .withColumn("available_tags", to_json("available_tags"))
        .withColumn("default_reaction_emoji", to_json("default_reaction_emoji"))
        .withColumn("source_keys", to_json("source_keys"))
        .write.format("postgresql")
        .mode("overwrite")
        .option("truncate", "true")
        .option("dbtable", "databricks_discord.channels")
        .option("host", postgres_host)
        .option("port", postgres_port)
        .option("database", postgres_db)
        .option("user", postgres_user)
        .option("password", postgres_password)
        .save()
    )


def overwrite_messages_table(spark: SparkSession, catalog: str, postgres_password: str):
    messages = full_table_name(
        catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_messages],
    )
    (
        spark.table(messages)
        .drop("content")
        .withColumn("reference", to_json("reference"))
        .withColumn("author", to_json("author"))
        .withColumn("mentions", to_json("mentions"))
        .withColumn("mention_channels", to_json("mention_channels"))
        .withColumn("embeds", to_json("embeds"))
        .withColumn("reactions", to_json("reactions"))
        .withColumn("thread", to_json("thread"))
        .withColumn("sticker_items", to_json("sticker_items"))
        .withColumn("poll", to_json("poll"))
        .withColumn("call", to_json("call"))
        .withColumn("source_keys", to_json("source_keys"))
        .write.format("postgresql")
        .mode("overwrite")
        .option("truncate", "true")
        .option("dbtable", "databricks_discord.messages")
        .option("host", postgres_host)
        .option("port", postgres_port)
        .option("database", postgres_db)
        .option("user", postgres_user)
        .option("password", postgres_password)
        .save()
    )


def overwrite_reactions_table(
    spark: SparkSession, catalog: str, postgres_password: str
):
    reactions = full_table_name(
        catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_reactions],
    )
    (
        spark.table(reactions)
        .withColumn("user", to_json("user"))
        .withColumn("emoji", to_json("emoji"))
        .withColumn("source_keys", to_json("source_keys"))
        .write.format("postgresql")
        .mode("overwrite")
        .option("truncate", "true")
        .option("dbtable", "databricks_discord.reactions")
        .option("host", postgres_host)
        .option("port", postgres_port)
        .option("database", postgres_db)
        .option("user", postgres_user)
        .option("password", postgres_password)
        .save()
    )


def overwrite_members_table(spark: SparkSession, catalog: str, postgres_password: str):
    members = full_table_name(
        catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_members],
    )
    (
        spark.table(members)
        .withColumn("user", to_json("user"))
        .withColumn("avatar_decoration_data", to_json("avatar_decoration_data"))
        .withColumn("source_keys", to_json("source_keys"))
        .write.format("postgresql")
        .mode("overwrite")
        .option("truncate", "true")
        .option("dbtable", "databricks_discord.members")
        .option("host", postgres_host)
        .option("port", postgres_port)
        .option("database", postgres_db)
        .option("user", postgres_user)
        .option("password", postgres_password)
        .save()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=[
            arguments.catalogs[arguments.catalogs_dev],
            arguments.catalogs[arguments.catalogs_hive],
        ],
        help="The unit catalog to use",
    )

    args = parser.parse_args()
    if args.target_catalog == arguments.catalogs[arguments.catalogs_dev]:
        return  # do not run dev version of this job

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())
    postgres_password = get_postgres_password(spark)

    overwrite_guilds_table(spark, args.target_catalog, postgres_password)
    overwrite_channels_table(spark, args.target_catalog, postgres_password)
    overwrite_messages_table(spark, args.target_catalog, postgres_password)
    overwrite_reactions_table(spark, args.target_catalog, postgres_password)
    overwrite_members_table(spark, args.target_catalog, postgres_password)


if __name__ == "__main__":
    main()
