import argparse
from typing import cast, Callable

from delta.tables import DeltaTable
from pyspark.sql import SparkSession
from pyspark.sql import DataFrame

from trend_etl.functions.index import full_table_name, get_silver_table
from trend_etl.transform.community_groups import from_bronze_get_community_groups_df
from trend_etl.functions.index import community_group_upsert_to_delta
import trend_etl.schema.args as arguments
from trend_etl.schema.types import Merge_Strategy

program = "load_community_groups"


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--idents-table",
        dest="idents_table",
        required=True,
        choices=arguments.tables,
        help="The source *_idents table in the selected schema",
    )

    parser.add_argument(
        "--data-source",
        dest="data_source",
        required=True,
        choices=arguments.data_sources,
        help="The social media data source",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        args.idents_table,
    )
    streaming_idents_table_df = spark.readStream.format("delta").table(idents_table)

    community_groups_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        get_silver_table(args.target_catalog, arguments.tables_community_groups),
    )
    checkpoint_loc = f"{args.checkpoints_path_prefix}/{arguments.tables[arguments.tables_community_groups]}/{args.idents_table}"
    delta_table = DeltaTable.forName(spark, community_groups_table)

    (
        from_bronze_get_community_groups_df(streaming_idents_table_df, args.data_source)
        .writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: community_group_upsert_to_delta(
                delta_table, batch_df
            )
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


if __name__ == "__main__":
    main()
