import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession

from trend_etl.functions.index import (
    full_table_name,
    daily_ingestion_counts_upsert_to_delta,
)
from trend_etl.transform.ingestion_counts import from_bronze_get_daily_ingestion_count
import trend_etl.schema.args as arguments

program = "load_daily_ingestion_counts"


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--bronze-catalog",
        dest="bronze_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--bronze-table",
        dest="bronze_table",
        required=True,
        choices=arguments.tables,
        help="The bronze table in the selected schema",
    )

    parser.add_argument(
        "--idents-table",
        dest="idents_table",
        required=True,
        choices=arguments.tables,
        help="The idents table in the selected schema",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    parser.add_argument(
        "--group-source-id-json-path",
        dest="group_source_id_json_path",
        required=True,
        help="The json path to the source's group id",
    )

    parser.add_argument(
        "--statement-type-index",
        dest="statement_type_index",
        type=int,
        required=True,
        choices=[
            arguments.statement_type_group,
            arguments.statement_type_subreddit,
            arguments.statement_type_post,
            arguments.statement_type_comment,
            arguments.statement_type_submission,
            arguments.statement_type_reply,
        ],
        help="Statement type index",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    bronze_table = full_table_name(
        args.bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        args.bronze_table,
    )
    group_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        args.idents_table,
    )
    daily_ingestion_counts_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_daily_ingestion_counts],
    )

    checkpoint_loc = (
        f"{args.checkpoints_path_prefix}"
        "/"
        f"{arguments.tables[arguments.tables_daily_ingestion_counts]}"
        "/"
        f"{args.bronze_table}"
    )
    streaming_bronze = spark.readStream.format("delta").table(bronze_table)
    group_idents = spark.read.format("delta").table(group_idents_table)
    delta_table = DeltaTable.forName(spark, daily_ingestion_counts_table)

    (
        from_bronze_get_daily_ingestion_count(
            streaming_bronze,
            group_idents,
            args.group_source_id_json_path,
            args.statement_type_index,
        )
        .writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: daily_ingestion_counts_upsert_to_delta(
                delta_table, batch_df, batch_id
            )
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("update")
        .trigger(availableNow=True)
        .start()
    )


if __name__ == "__main__":
    main()
