import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession

from trend_etl.functions.index import (
    full_table_name,
    reactions_upsert_to_delta,
    get_silver_table,
)
from trend_etl.transform.reactions import from_bronze_get_reactions_df
import trend_etl.schema.args as arguments

program = "load_reactions"


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--bronze-catalog",
        dest="bronze_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    kafka_reactions_table = full_table_name(
        args.bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_reactions],
    )
    streaming_kafka_reactions = spark.readStream.format("delta").table(
        kafka_reactions_table
    )

    message_idents_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_message_idents],
    )
    message_idents = spark.read.format("delta").table(message_idents_table)

    reaction_table = full_table_name(
        args.target_catalog,
        arguments.schemas[arguments.schemas_silver],
        get_silver_table(args.target_catalog, arguments.tables_discord_reactions),
    )
    checkpoint_loc = f"{args.checkpoints_path_prefix}/{arguments.tables[arguments.tables_discord_reactions]}"
    delta_table = DeltaTable.forName(spark, reaction_table)

    (
        from_bronze_get_reactions_df(streaming_kafka_reactions, message_idents)
        .writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: reactions_upsert_to_delta(delta_table, batch_df)
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


if __name__ == "__main__":
    main()
