import argparse
from typing import cast

from pyspark.sql import SparkSession

from trend_etl.functions.index import full_table_name
from trend_etl.transform.replies import from_kafka_get_bronze_replies_df
import trend_etl.schema.args as arguments

program = "read_kafka_replies"


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--kafka-broker",
        dest="kafka_broker",
        required=True,
        help="The host:port for the kafka broker",
    )

    parser.add_argument(
        "--target-topic",
        dest="target_topic",
        required=True,
        choices=arguments.kafka_topics,
        help="The kafka topic to subscribe to for updates",
    )

    parser.add_argument(
        "--bronze-catalog",
        dest="bronze_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--bronze-table",
        dest="bronze_table",
        required=True,
        choices=arguments.tables,
        help="The table in the selected schema",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())
    args = parser.parse_args()
    bronze_table = full_table_name(
        args.bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        args.bronze_table,
    )
    checkpoint_loc = (
        f"{args.checkpoints_path_prefix}/{args.bronze_table}/{args.target_topic}"
    )

    bronze_df = from_kafka_get_bronze_replies_df(
        spark.readStream.format("kafka")
        .option("kafka.bootstrap.servers", args.kafka_broker)
        .option("subscribe", args.target_topic)
        .option("startingOffsets", "earliest")
        .option("failOnDataLoss", "false")
        .load(),
        args.target_topic,
    )

    (
        bronze_df.writeStream.outputMode("append")
        .option("checkpointLocation", checkpoint_loc)
        .trigger(availableNow=True)
        .toTable(bronze_table)
    )


if __name__ == "__main__":
    main()
