import argparse
from typing import cast
import json

from slack_sdk import WebClient

from pyspark.dbutils import DBUtils  # type: ignore
from delta.tables import DeltaTable
from pyspark.sql import SparkSession
from pyspark.sql.types import LongType
from pyspark.sql.functions import (
    col,
    lit,
    concat_ws,
    format_number,
    date_format,
    sum,
    last,
)

program = "report_statements"


def report_message_for_notifier(
    spark: SparkSession,
    slack_client: WebClient,
    slack_channel_id: str,
    table_notifier: dict,
    job_run_id: str,
) -> None:
    delta_table = DeltaTable.forName(spark, table_notifier["table"])
    message = concat_ws(
        " ",
        lit(table_notifier["notification_label"]),
        lit("inserted"),
        col("total_rows_inserted"),
        lit("statements on"),
        date_format("timestamp", "MMM-dd-y KK:mm a (O)"),
    )
    history_rows = delta_table.history().where(col("job.jobRunId") == lit(job_run_id))

    if history_rows.count() == 0:
        slack_client.chat_postMessage(
            channel=slack_channel_id,
            text=f"No new {table_notifier['notification_label']} inserted",
        )
    else:
        history_row = (
            history_rows.select(
                col("timestamp"),
                col("operationMetrics.numTargetRowsInserted")
                .cast(LongType())
                .alias("num_inserted"),
            )
            .agg(
                last("timestamp").alias("timestamp"),
                format_number(sum(col("num_inserted")), 0).alias("total_rows_inserted"),
            )
            .select(message.alias("m"))
            .collect()
        )
        slack_client.chat_postMessage(
            channel=slack_channel_id, text=history_row[0]["m"]
        )


def get_slack_api_token(spark: SparkSession) -> str:
    dbutils = DBUtils(spark)

    token = dbutils.secrets.get(scope="slack", key="trendy")

    return token


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    # JSON schema {table:string, notification_label:string}
    parser.add_argument(
        "--table-notifier",
        dest="table_notifier",
        required=True,
        help="A JSON str that is a table notifiers for reporting",
    )

    parser.add_argument(
        "--job-run-id",
        dest="job_run_id",
        required=True,
        help="The unique databricks (job) run id that runs this",
    )

    parser.add_argument(
        "--slack-channel-id",
        dest="slack_channel_id",
        required=True,
        help="The unique databricks (job) run id that runs this",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    table_notifier = json.loads(args.table_notifier)

    trendy_token = get_slack_api_token(spark)
    slack_client = WebClient(token=trendy_token)

    report_message_for_notifier(
        spark, slack_client, args.slack_channel_id, table_notifier, args.job_run_id
    )


if __name__ == "__main__":
    main()
