import os
import json

from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType, BinaryType
from pyspark.sql.functions import col, lit, current_timestamp, udf

import trend_etl.schema.fields as fields

script_dir = os.path.dirname(__file__)


@udf(returnType=BinaryType())
def str_to_binary(value: str, encoding: str) -> bytes:
    return value.encode(encoding)


def getspark() -> SparkSession:
    return SparkSession.builder.master("local").appName("local tests").getOrCreate()


def schema_df(*pathcomps):
    with open(os.path.join(script_dir, *pathcomps), "r", encoding="utf-8") as file:
        return StructType.fromJson(json.loads(file.read()))


def get_df(spark, schema: str, *pathcomps) -> DataFrame:
    return spark.read.schema(schema).json(os.path.join(script_dir, *pathcomps))


def get_kafka_df(spark, *pathcomps) -> DataFrame:
    return spark.read.text(os.path.join(script_dir, *pathcomps)).select(
        str_to_binary(col("value"), lit("utf-8")).alias(fields.kafka_value_field),
        current_timestamp().alias(fields.kafka_timestamp_field),
    )


def get_archive_df(spark, *pathcomps) -> DataFrame:
    return spark.read.text(os.path.join(script_dir, *pathcomps)).select(
        col("value").alias(fields.archive_value_field),
        lit("62e79d1a455bff37d7d29528").alias(fields.archive_job_id_field),
    )


def assert_dataframes_equal(df1: DataFrame, df2: DataFrame):
    """
    Assert that two DataFrames are equal.

    Parameters:
    df1 (DataFrame): The first DataFrame to compare.
    df2 (DataFrame): The second DataFrame to compare.
    ignore_nullable (bool): If True, ignores nullable differences in schema.

    Raises:
    AssertionError: If the DataFrames are not equal.
    """
    # Check schema equality
    schema1 = df1.schema
    schema2 = df2.schema

    assert schema1 == schema2, f"Schema mismatch: {schema1} != {schema2}"

    # Check row count
    count1 = df1.count()
    count2 = df2.count()
    assert count1 == count2, f"Row count mismatch: {count1} != {count2}"

    # Check content equality
    diff = df1.subtract(df2).union(df2.subtract(df1)).count()
    assert diff == 0, f"DataFrame content mismatch: {diff} rows differ"
