import unittest
from unittest.mock import Mock

from pyspark.sql import SparkSession
from pyspark.sql.functions import reverse
from trend_etl.test_utils import getspark, get_kafka_df, get_archive_df
from trend_etl.transform.groups import (
    from_kafka_get_bronze_groups_df,
    from_archive_get_bronze_groups_df,
)
from trend_etl.transform.posts import (
    from_kafka_get_bronze_posts_df,
    from_archive_get_bronze_posts_df,
)
from trend_etl.transform.comments import (
    from_kafka_get_bronze_comments_df,
    from_archive_get_bronze_comments_df,
)
from trend_etl.transform.subreddits import (
    from_kafka_get_bronze_subreddits_df,
    from_archive_get_bronze_subreddits_df,
)
from trend_etl.transform.submissions import (
    from_kafka_get_bronze_submissions_df,
    from_reddit_archive_get_bronze_submissions_df,
    from_pushshift_archive_get_bronze_submissions_df,
)
from trend_etl.transform.replies import (
    from_kafka_get_bronze_replies_df,
    from_reddit_archive_get_bronze_replies_df,
    from_pushshift_archive_get_bronze_replies_df,
)
from trend_etl.transform.idents import from_bronze_get_idents_df
from trend_etl.transform.guilds import from_kafka_get_bronze_guilds_df
from trend_etl.transform.channels import from_kafka_get_bronze_channels_df
from trend_etl.transform.messages import from_kafka_get_bronze_messages_df
from trend_etl.transform.reactions import from_kafka_get_bronze_reactions_df
from trend_etl.functions.index import mongo_objectid_is_valid
import trend_etl.schema.fields as fields
import trend_etl.schema.args as arguments


class TestBronzeTranforms(unittest.TestCase):
    spark: SparkSession

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def test_mongodb_is_valid(self) -> None:
        data = [("apple", "62fd05f687acfd8240b5de3a"), ("lemon", "00000000-old_posts")]
        got_df = self.__class__.spark.createDataFrame(
            data, schema="fruit string, job_id string"
        )

        got_length = got_df.where(mongo_objectid_is_valid("job_id")).count()
        want_length = 1
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_groups(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "groups.jsonl")
        got_df = from_kafka_get_bronze_subreddits_df(source_df, "groups")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_subreddits(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "subreddits.jsonl")
        got_df = from_kafka_get_bronze_groups_df(source_df, "subreddits")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_posts(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "posts.jsonl")
        got_df = from_kafka_get_bronze_posts_df(source_df, "posts", reverse)

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_comments(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "comments.jsonl")
        got_df = from_kafka_get_bronze_comments_df(source_df, "comments", reverse)

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_submissions(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "submissions.jsonl")
        got_df = from_kafka_get_bronze_submissions_df(source_df, "submissions")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_replies(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "submissions.jsonl")
        got_df = from_kafka_get_bronze_replies_df(source_df, "replies")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_archive_get_groups(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "groups.jsonl")
        group_test_id = "100"  # "200" is not a test statment

        got_df = from_archive_get_bronze_groups_df(archive_df, [group_test_id], False)
        self.assertEqual(1, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_df = from_archive_get_bronze_groups_df(archive_df, [group_test_id], True)
        self.assertEqual(1, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_subreddits(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "subreddits.jsonl"
        )

        got_df = from_archive_get_bronze_subreddits_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_posts(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "posts.jsonl")
        group_test_id = "100"  # "200" is not a test statment

        got_df = from_archive_get_bronze_posts_df(
            archive_df, [group_test_id], False, reverse
        )
        self.assertEqual(1, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_df = from_archive_get_bronze_posts_df(
            archive_df, [group_test_id], True, reverse
        )
        self.assertEqual(
            archive_df.count() - 1, got_df.count(), "dataframe size different"
        )

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_comments(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "comments.jsonl")
        group_test_id = "100"  # "200" is not a test statment

        got_df = from_archive_get_bronze_comments_df(
            archive_df, [group_test_id], False, reverse
        )
        self.assertEqual(4, got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_df = from_archive_get_bronze_comments_df(
            archive_df, [group_test_id], True, reverse
        )
        self.assertEqual(
            archive_df.count() - 4, got_df.count(), "dataframe size different"
        )

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_submissions_reddit(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "submissions.jsonl"
        )

        got_df = from_reddit_archive_get_bronze_submissions_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_submissions_pushshift(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "submissions.pushshift.jsonl"
        )

        got_df = from_pushshift_archive_get_bronze_submissions_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_replies_reddit(self) -> None:
        archive_df = get_archive_df(self.__class__.spark, "test_data", "replies.jsonl")

        got_df = from_reddit_archive_get_bronze_replies_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_archive_get_replies_pushshift(self) -> None:
        archive_df = get_archive_df(
            self.__class__.spark, "test_data", "replies.pushshift.jsonl"
        )

        got_df = from_pushshift_archive_get_bronze_replies_df(archive_df)
        self.assertEqual(archive_df.count(), got_df.count(), "dataframe size different")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

    def test_facebook_get_idents_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "groups.jsonl")
        bronze_table = from_kafka_get_bronze_groups_df(source_df, "groups")
        got_df = from_bronze_get_idents_df(
            bronze_table, "$.id", arguments.tables[arguments.tables_group_idents]
        )

        got_columns = got_df.columns
        want_columns = [
            fields.idents_group_id_field,
            fields.idents_source_id_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_reddit_get_idents_df(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "subreddits.jsonl")
        bronze_table = from_kafka_get_bronze_subreddits_df(source_df, "unittest")
        got_df = from_bronze_get_idents_df(
            bronze_table,
            "$.data.name",
            arguments.tables[arguments.tables_subreddit_idents],
        )

        got_columns = got_df.columns
        want_columns = [
            fields.idents_group_id_field,
            fields.idents_source_id_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_guilds(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "guilds.jsonl")
        got_df = from_kafka_get_bronze_guilds_df(source_df, "guilds")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_channels(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "channels.jsonl")
        got_df = from_kafka_get_bronze_channels_df(source_df, "channels")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_messages(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "messages.jsonl")
        got_df = from_kafka_get_bronze_messages_df(source_df, "messages")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.created_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")

    def test_kafka_get_reactions(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "reactions.jsonl")
        got_df = from_kafka_get_bronze_reactions_df(source_df, "reactions")

        got_columns = got_df.columns
        want_columns = [
            fields.bronze_value_field,
            fields.bronze_provenance_field,
            fields.bronze_timestamp_field,
            fields.collected_day_field,
        ]

        self.assertEqual(want_columns, got_columns, "columns match error")

        got_length = got_df.count()
        want_length = source_df.count()
        self.assertEqual(want_length, got_length, "dataframe size different")
