import unittest
from unittest.mock import patch
from datetime import datetime

from pyspark.sql import SparkSession
from pyspark.sql.functions import col, lit

from trend_etl.test_utils import getspark, get_kafka_df, get_df
from trend_etl.functions.index import conv_timestamp
from trend_etl.transform.discord_stats import (
    total_journeys_number,
    journey_conversion_number,
    total_statements_number,
    active_users_for_threads,
    active_users_for_messages,
    active_users_for_reactions,
    total_registered_users_number,
    active_users_days_ago_number,
    active_audience_days_ago_number,
    retention_audience_number,
    days_ago_cond,
    greater_days_cond,
)
from trend_etl.transform.channels import (
    from_bronze_get_channels_df,
    from_kafka_get_bronze_channels_df,
)
from trend_etl.transform.messages import (
    from_bronze_get_messages_df,
    from_kafka_get_bronze_messages_df,
)
from trend_etl.transform.reactions import (
    from_bronze_get_reactions_df,
    from_kafka_get_bronze_reactions_df,
)
from trend_etl.transform.members import (
    from_bronze_get_members_df,
    from_kafka_get_bronze_members_df,
)
import trend_etl.schema.args as arguments


class TestSilverTranforms(unittest.TestCase):
    spark: SparkSession

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def test_total_journeys_number(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "channels.jsonl")
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        channels_table = from_bronze_get_channels_df(
            from_kafka_get_bronze_channels_df(source_df, "channels"),
            guild_idents,
            channel_idents,
        )

        self.assertEqual(
            1,
            total_journeys_number(channels_table, ["1215765928013201488"]),
            "Expected one non-trend employee to haved posted a journey",
        )

    def test_journeys_conversion_number(self) -> None:
        source_channels = get_kafka_df(
            self.__class__.spark, "test_data", "channels.jsonl"
        )
        source_members = get_kafka_df(
            self.__class__.spark, "test_data", "members.jsonl"
        )
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        channels_table = from_bronze_get_channels_df(
            from_kafka_get_bronze_channels_df(source_channels, "channels"),
            guild_idents,
            channel_idents,
        )
        guild_members = from_bronze_get_members_df(
            from_kafka_get_bronze_members_df(source_members, "members"), guild_idents
        )

        self.assertEqual(
            1,
            journey_conversion_number(
                channels_table, guild_members, ["1215765928013201488"]
            ),
            "Expected one non-trend employee to haved posted a journey",
        )

    def test_total_statements_number(self) -> None:
        source_channels = get_kafka_df(
            self.__class__.spark, "test_data", "channels.jsonl"
        )
        source_messages = get_kafka_df(
            self.__class__.spark, "test_data", "messages.jsonl"
        )
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        message_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "message_idents.jsonl",
        )

        threads_table = from_bronze_get_channels_df(
            from_kafka_get_bronze_channels_df(source_channels, "channels"),
            guild_idents,
            channel_idents,
        ).where(col("type").isin(arguments.discord_thread_types))

        messages_table = from_bronze_get_messages_df(
            from_kafka_get_bronze_messages_df(source_messages, "messages"),
            guild_idents,
            channel_idents,
            message_idents,
        )

        self.assertEqual(
            20,
            total_statements_number(threads_table, messages_table),
            "Expected to have 20 threads + messages not authored by trend employees",
        )

    @patch("trend_etl.transform.discord_stats.current_timestamp")
    def test_active_users_for_threads(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 8))
        mock_current_timestamp.return_value = fixed_current_timestamp

        source_channels = get_kafka_df(
            self.__class__.spark, "test_data", "channels.jsonl"
        )
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )

        threads_table = from_bronze_get_channels_df(
            from_kafka_get_bronze_channels_df(source_channels, "channels"),
            guild_idents,
            channel_idents,
        ).where(col("type").isin(arguments.discord_thread_types))

        thread_created_column = conv_timestamp("thread_metadata.create_timestamp")
        threads_date_cond = days_ago_cond(thread_created_column, 5)
        active_users = active_users_for_threads(threads_table, threads_date_cond)

        self.assertEqual(
            active_users.count(),
            1,
            "Expects one thread created by a non-trend employee",
        )

    @patch("trend_etl.transform.discord_stats.current_timestamp")
    def test_active_users_for_messages(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        source_messages = get_kafka_df(
            self.__class__.spark, "test_data", "messages.jsonl"
        )
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        message_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "message_idents.jsonl",
        )

        messages_table = from_bronze_get_messages_df(
            from_kafka_get_bronze_messages_df(source_messages, "messages"),
            guild_idents,
            channel_idents,
            message_idents,
        )

        messages_date_cond = days_ago_cond(col("created_at"), 5)
        active_users = active_users_for_messages(messages_table, messages_date_cond)

        # NOTE: The 2 messages created on 2024-04-03 are older than the time filter
        self.assertEqual(
            active_users.count(),
            15,
            "Expects 15 messages created by non-trend employees",
        )

        messages_date_cond = greater_days_cond(col("created_at"), 5)
        active_users = active_users_for_messages(messages_table, messages_date_cond)
        self.assertEqual(
            active_users.count(),
            4,
            "Expects 4 messages created by non-trend employees",
        )

    @patch("trend_etl.transform.discord_stats.current_timestamp")
    def test_active_users_for_reactions(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        source_messages = get_kafka_df(
            self.__class__.spark, "test_data", "messages.jsonl"
        )
        source_reactions = get_kafka_df(
            self.__class__.spark, "test_data", "reactions.jsonl"
        )

        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        message_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "message_idents.jsonl",
        )

        messages_table = from_bronze_get_messages_df(
            from_kafka_get_bronze_messages_df(source_messages, "messages"),
            guild_idents,
            channel_idents,
            message_idents,
        )
        reactions_table = from_bronze_get_reactions_df(
            from_kafka_get_bronze_reactions_df(source_reactions, "reactions"),
            message_idents,
        )

        messages_date_cond = days_ago_cond(col("created_at"), 5)
        active_users = active_users_for_reactions(
            messages_table, reactions_table, messages_date_cond
        )

        self.assertEqual(
            active_users.count(),
            5,
            "Expects 5 reactions created by non-trend employees",
        )

    def test_total_registered_users_number(self) -> None:
        source_df = get_kafka_df(self.__class__.spark, "test_data", "members.jsonl")
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        guild_members = from_bronze_get_members_df(
            from_kafka_get_bronze_members_df(source_df, "members"), guild_idents
        )

        self.assertEqual(
            4,
            total_registered_users_number(guild_members),
            "Expected 4 non-trend employees as members",
        )

    @patch("trend_etl.transform.discord_stats.current_timestamp")
    def test_active_users_days_ago_number(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        source_channels = get_kafka_df(
            self.__class__.spark, "test_data", "channels.jsonl"
        )
        source_messages = get_kafka_df(
            self.__class__.spark, "test_data", "messages.jsonl"
        )
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        message_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "message_idents.jsonl",
        )

        threads_table = from_bronze_get_channels_df(
            from_kafka_get_bronze_channels_df(source_channels, "channels"),
            guild_idents,
            channel_idents,
        ).where(col("type").isin(arguments.discord_thread_types))
        messages_table = from_bronze_get_messages_df(
            from_kafka_get_bronze_messages_df(source_messages, "messages"),
            guild_idents,
            channel_idents,
            message_idents,
        )

        active_users = active_users_days_ago_number(threads_table, messages_table, 5, 1)

        self.assertEqual(
            active_users,
            2,
            "Expects 2 active users using message_count_threshold=1",
        )

        active_users = active_users_days_ago_number(threads_table, messages_table, 5, 4)

        self.assertEqual(
            active_users,
            1,
            "Expects 1 active users using message_count_threshold=4",
        )

    @patch("trend_etl.transform.discord_stats.current_timestamp")
    def test_active_audience_days_ago_number(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        source_channels = get_kafka_df(
            self.__class__.spark, "test_data", "channels.jsonl"
        )
        source_messages = get_kafka_df(
            self.__class__.spark, "test_data", "messages.jsonl"
        )
        source_reactions = get_kafka_df(
            self.__class__.spark, "test_data", "reactions.jsonl"
        )
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        message_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "message_idents.jsonl",
        )

        threads_table = from_bronze_get_channels_df(
            from_kafka_get_bronze_channels_df(source_channels, "channels"),
            guild_idents,
            channel_idents,
        ).where(col("type").isin(arguments.discord_thread_types))
        messages_table = from_bronze_get_messages_df(
            from_kafka_get_bronze_messages_df(source_messages, "messages"),
            guild_idents,
            channel_idents,
            message_idents,
        )
        reactions_table = from_bronze_get_reactions_df(
            from_kafka_get_bronze_reactions_df(source_reactions, "reactions"),
            message_idents,
        )

        active_audience = active_audience_days_ago_number(
            threads_table, messages_table, reactions_table, 5, 1, 1
        )

        self.assertEqual(
            active_audience,
            2,
            "Expects 2 active audience",
        )

    @patch("trend_etl.transform.discord_stats.current_timestamp")
    def test_retention_audience_number_number(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        source_channels = get_kafka_df(
            self.__class__.spark, "test_data", "channels.jsonl"
        )
        source_messages = get_kafka_df(
            self.__class__.spark, "test_data", "messages.jsonl"
        )
        source_reactions = get_kafka_df(
            self.__class__.spark, "test_data", "reactions.jsonl"
        )
        guild_idents = get_df(
            self.__class__.spark,
            "group_id string, id string",
            "test_data",
            "guild_idents.jsonl",
        )
        channel_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "channel_idents.jsonl",
        )
        message_idents = get_df(
            self.__class__.spark,
            "doc_uuid string, id string",
            "test_data",
            "message_idents.jsonl",
        )

        threads_table = from_bronze_get_channels_df(
            from_kafka_get_bronze_channels_df(source_channels, "channels"),
            guild_idents,
            channel_idents,
        ).where(col("type").isin(arguments.discord_thread_types))
        messages_table = from_bronze_get_messages_df(
            from_kafka_get_bronze_messages_df(source_messages, "messages"),
            guild_idents,
            channel_idents,
            message_idents,
        )
        reactions_table = from_bronze_get_reactions_df(
            from_kafka_get_bronze_reactions_df(source_reactions, "reactions"),
            message_idents,
        )

        retention_audience = retention_audience_number(
            threads_table, messages_table, reactions_table, 5
        )

        self.assertEqual(
            retention_audience,
            1,
            "Expects 1 for audience retention",
        )
