from pyspark.sql import DataFrame

from pyspark.sql.functions import col, lit, to_json, struct, format_string, when
from pyspark.sql.types import (
    StringType,
    BinaryType,
    TimestampType,
    LongType,
    DoubleType,
)

import trend_etl.schema.fields as fields
import trend_etl.schema.args as arguments
from trend_etl.functions.index import (
    jobid_2_timestamp,
    parse_reddit_replies,
    parse_pushshift_replies,
    get_created_day,
    is_scientific,
)


def from_reddit_archive_get_bronze_replies_df(archive_df: DataFrame) -> DataFrame:
    parsed = archive_df.withColumn(
        fields.archive_value_field,
        parse_reddit_replies(col(fields.archive_value_field)),
    )
    created_utc_field = col(f"{fields.archive_value_field}.data.created_utc")

    return parsed.select(
        to_json(col(fields.archive_value_field))
        .cast(BinaryType())
        .alias(fields.bronze_value_field),
        lit("archive/reddit/replies").alias(fields.bronze_provenance_field),
        jobid_2_timestamp(fields.archive_job_id_field).alias(
            fields.bronze_timestamp_field
        ),
        get_created_day(created_utc_field.cast(TimestampType())).alias(
            fields.created_day_field
        ),
    )


def from_pushshift_archive_get_bronze_replies_df(
    archive_df: DataFrame,
) -> DataFrame:
    parsed = archive_df.withColumn(
        fields.archive_value_field,
        parse_pushshift_replies(col(fields.archive_value_field)),
    )
    id_field = col(f"{fields.archive_value_field}.id")
    created_utc_field = col(f"{fields.archive_value_field}.created_utc")

    return (
        parsed.withColumn(
            fields.archive_value_field,
            col(fields.archive_value_field).withField(
                "name",
                format_string("t1_%s", id_field),
            ),
        )
        .withColumn(
            fields.archive_value_field,
            col(fields.archive_value_field).withField(
                "created_utc",
                created_utc_field.cast(LongType()),
            ),
        )
        .select(
            to_json(
                struct(
                    lit("t1").alias("kind"),
                    col(fields.archive_value_field).alias("data"),
                    lit("pushshift_archive").alias("note"),
                )
            )
            .cast(BinaryType())
            .alias(fields.bronze_value_field),
            lit("archive/pushshift/replies").alias(fields.bronze_provenance_field),
            jobid_2_timestamp(fields.archive_job_id_field).alias(
                fields.bronze_timestamp_field
            ),
            get_created_day(created_utc_field.cast(TimestampType())).alias(
                fields.created_day_field
            ),
        )
    )


# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables (for all reddit data kinds)
def from_kafka_get_bronze_replies_df(input_df: DataFrame, target_topic: str):
    parsed = input_df.withColumn(
        fields.kafka_value_field,
        parse_reddit_replies(col(fields.kafka_value_field).cast(StringType())),
    )
    created_utc_field = col(f"{fields.archive_value_field}.data.created_utc")

    return parsed.select(
        to_json(col(fields.kafka_value_field))
        .cast(BinaryType())
        .alias(fields.bronze_value_field),
        lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_created_day(created_utc_field.cast(TimestampType())).alias(
            fields.created_day_field
        ),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   subreddit_idents: subreddit uuid idents table
#   submission_idents: submission uuid idents table
#   reply_idents: reply uuid idents table
#   bot_authors: reddit bot users table
# return: dataframe in the schema of `reddit_replies`
def from_bronze_get_replies_df(
    streaming_bronze: DataFrame,
    subreddit_idents: DataFrame,
    submission_idents: DataFrame,
    reply_idents: DataFrame,
    bot_authors: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reddit_replies(col(fields.bronze_value_field).cast(StringType())).alias(
            "value"
        ),
        col(fields.bronze_timestamp_field),
    )
    flattened_df = parsed_df.select(
        col("value.data.*"),
        col(fields.bronze_timestamp_field),
    )

    # filter out bot submissions
    bot_filtered_df = flattened_df.join(
        bot_authors, fields.reddit_bot_users_author, "left_anti"
    )

    test_filtered_df = bot_filtered_df.where(
        col("subreddit_id").isin(arguments.test_subreddits) == False
    )

    # NOTE: left outer joins (below) will cause any nulls to be
    #       detected as a downstream error.

    # join with subreddit_idents
    subreddits_df = test_filtered_df.join(
        subreddit_idents.withColumnRenamed(
            fields.idents_source_id_field, "subreddit_id"
        ),
        "subreddit_id",
        "inner",  # filter out replies from subreddit_type = 'user' and like which we don't track
    ).drop("subreddit_id")

    # join with reply_idents to get doc_uuid
    reply_idents_df = subreddits_df.join(
        reply_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
        "name",
        "left_outer",
    ).withColumnRenamed("name", "_source_key_id")

    # join with submission_idents to get corresponding submission uuid fk
    assoc_submission_df = reply_idents_df.join(
        submission_idents.withColumnsRenamed(
            {
                fields.idents_doc_id_field: fields.reddit_replies_submission_uuid,
                fields.idents_source_id_field: "link_id",
            }
        ),
        "link_id",
        "left_outer",
    ).withColumnRenamed("link_id", "_source_key_link_id")

    # join with comment_idents to get parent uuid fk
    assoc_parent_df = assoc_submission_df.withColumnRenamed(
        "parent_id", "_source_key_parent_id"
    ).join(
        reply_idents.withColumnsRenamed(
            {
                fields.idents_doc_id_field: fields.reddit_replies_parent_uuid,
                fields.idents_source_id_field: "_source_key_parent_id",
            }
        ),
        "_source_key_parent_id",
        "left_outer",
    )

    return (
        assoc_parent_df.withColumn(
            fields.reddit_replies_created_at,
            when(
                is_scientific(col("created_utc")),
                col("created_utc").cast(DoubleType()).cast(LongType()),
            )
            .otherwise(col("created_utc").cast(LongType()))
            .cast(TimestampType()),
        )
        .drop("created_utc")
        .withColumnRenamed("body", fields.reddit_replies_text)
        .withColumn(
            fields.source_keys_field,
            struct(
                col("_source_key_id").alias("name"),
                col("_source_key_link_id").alias("submission_id"),
                col("_source_key_parent_id").alias("parent_id"),
            ),
        )
        .drop("_source_key_id", "_source_key_link_id", "_source_key_parent_id")
    )
