import spacy  # type: ignore
from spacy import displacy  # type: ignore
from spacy.language import Language  # type: ignore
from spacy.ml import Doc  # type: ignore
import pandas as pd
from typing import Optional
from pyspark.sql.types import StringType
from pyspark.sql.functions import pandas_udf

from trend_etl.schema.args import ner_model_path

# defer loading NLP until needed by worker(s)
NLP: Optional[dict[str, Language]] = None


# arguments:
#   model_path: dbfs path to finished NER model
# return: adds entry to global NLP, returns language model
def get_spacy_magic(model_name: str) -> Language:
    global NLP
    if NLP is None:
        NLP = {}
    if model_name not in NLP:
        spacy.require_gpu()
        NLP[model_name] = spacy.load(model_name)
    return NLP[model_name]


# arguments:
#   doc: spacy evaluation result
# return: text with names replaced with token
def anonymize_text(doc: Doc) -> Optional[str]:
    # if doc.text is None:
    #     return None
    replacement_token = "{{NAME}}"  # constant
    parts = []  # parts of text except person names identified
    start_idx = 0  # starting point for remaining text
    for ent in doc.ents:
        if ent.label_ == "PERSON":
            parts.append(doc.text[start_idx : ent.start_char])
            parts.append(
                replacement_token
            )  # replace person's name with `replacement_token` token
            start_idx = ent.end_char
    # test to see if any remaining text
    if start_idx < len(doc.text):
        parts.append(doc.text[start_idx:])
    return "".join(parts)


def none_to_empty_string(v):
    if v is None:
        return ""
    return v


def empty_string_to_none(v):
    if v == "":
        return None
    return v


# vectorized udf
# arguments:
#   s: pandas series of text
# return: pandas series of text
@pandas_udf(returnType=StringType())  # type: ignore
def apply_ner(s: pd.Series) -> pd.Series:
    nlp = get_spacy_magic(ner_model_path)
    fixed_series = s.apply(none_to_empty_string)
    result_series = pd.Series([anonymize_text(doc) for doc in nlp.pipe(fixed_series)])
    return result_series.apply(empty_string_to_none)
